/*
 *  Licensed to the Apache Software Foundation (ASF) under one
 *  or more contributor license agreements.  See the NOTICE file
 *  distributed with this work for additional information
 *  regarding copyright ownership.  The ASF licenses this file
 *  to you under the Apache License, Version 2.0 (the
 *  "License"); you may not use this file except in compliance
 *  with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *   * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 */
package org.wso2.registry.versions;

import org.wso2.registry.Resource;
import org.wso2.registry.RegistryConstants;
import org.wso2.registry.jdbc.dao.ResourceDAO;

import java.sql.*;
import java.util.ArrayList;

/**
 * This class is to handle all the version related stuff , putting into the dependecy tables ,
 * updating the tables and ect..
 */
public class VersionHandler {

    /**
     * To add a new resource or to update a new resource we can use this.
     *
     * @param connection : Connection to DB
     * @param resource   : Artifcat need to put
     * @throws SQLException : If something went wrong
     */
    public static void addResourceToVersionTable(Connection connection,
                                                 Resource resource) throws SQLException {
        if (resource.getLastModified() == null) {
            resource.setLastModified(new java.util.Date());
        }

        long versionNumber = insertToVersionTable(connection, resource.getContent(),
                resource.getId(),
                resource.getLastUpdaterUserName(),
                resource.getLastModified().getTime());
        //updating max version table
        updateDependency(resource, connection, versionNumber, false);
    }

    private static void updateDependency(Resource resource,
                                         Connection connection,
                                         long versionNumber,
                                         boolean delete) throws SQLException {
        //updating dependency table , so that we can rollback to a particular version
        //updating for the current resource
        long parentNextVersionNumber = updateDependencyTable(resource, connection, versionNumber, delete);
        // a/b/c/r1.wsdl
        //Next a/b/c
        // Next a/b
        //Next a
        if (parentNextVersionNumber != -1) {
            //updating its parent
            String path = resource.getPath();
            while (path != null && path.indexOf("/") > -1) {
                int lastIndex = path.lastIndexOf("/");
                String parentPath = path.substring(0, lastIndex);
                ResourceDAO resourceDAO = new ResourceDAO();
                Resource parentResource = resourceDAO.get(parentPath, connection);
                if (parentResource != null) {
                    path = parentResource.getPath();
                    parentNextVersionNumber = updateDependencyTable(
                            parentResource, connection, parentNextVersionNumber, false);
                    if (parentNextVersionNumber == -1) {
                        break;
                    }
                } else {
                    break;
                }
            }
        }
    }

    /**
     * To delete a resource , what actualy happen here is that update the version table with the new data
     * and resource will not actualy removed from the table
     *
     * @param resource   : Resource to be deleted
     * @param connection : Connection to the DB
     * @throws java.sql.SQLException : if something went wrong
     */
    public static void deleteResource(Resource resource,
                                      Connection connection) throws SQLException {
        long versionNumber = getLatestVerison(resource.getId(), connection);
        updateDependency(resource, connection, versionNumber, true);
    }

    /**
     * To get the version latest version number for a given artifact
     *
     * @param artifactId : Resource ID
     * @param connection : DB connection
     * @return : will retuen the version number , it could be either -1 or valid version number
     * @throws SQLException : If something went wrong
     */
    public static long getLatestVerison(long artifactId, Connection connection) throws SQLException {
        String sql = "SELECT MAX(VN) FROM VERSIONS WHERE AID=?";
        PreparedStatement s = connection.prepareStatement(sql);
        s.setLong(1, artifactId);
        ResultSet result = s.executeQuery();
        if (result.next()) {
            return result.getLong(1);
        }
        return -1;
    }

    public static String[] getVersions(long artifactId, Connection connection) throws SQLException {
        String SQL = "SELECT VN FROM VERSIONS WHERE AID=?";
        PreparedStatement s = connection.prepareStatement(SQL);
        s.setLong(1, artifactId);
        ResultSet result = s.executeQuery();
        ArrayList versionNumbers = new ArrayList();
        while (result.next()) {
            versionNumbers.add("" + result.getLong("VN"));
        }
        return (String[]) versionNumbers.toArray(new String[versionNumbers.size()]);
    }

    /**
     * To get the depndents path for a given artifact
     *
     * @param artifcatID : ID
     * @param connection : Connection to DB
     * @return : List of paths
     * @throws SQLException : If something went wrong
     */
    public static String[] getChildrenPath(long artifcatID,
                                           Connection connection) throws SQLException {
        long versionNumber = getLatestVerison(artifcatID, connection);
        if (versionNumber == -1) {
            return null;
        }
        String selectSql = "SELECT DAID , DVN FROM DEPENDENCY WHERE AID=? AND VN=?";
        PreparedStatement s = connection.prepareStatement(selectSql);
        s.setLong(1, artifcatID);
        s.setLong(2, versionNumber);
        ResultSet result = s.executeQuery();

        ArrayList dependeyPathMap = new ArrayList();

        //to see whether the patent node is there in the table , if not need to add that
        while (result.next()) {
            long daid = result.getLong("DAID");
            String SQL = "SELECT PATH FROM ARTIFACTS WHERE AID=?";
            PreparedStatement ps = connection.prepareStatement(SQL);
            ps.setLong(1, daid);
            ResultSet pathResult = ps.executeQuery();
            if (pathResult.next()) {
                String path = result.getString("PATH");
                dependeyPathMap.add(path);
            }
        }
        return (String[]) (dependeyPathMap.toArray(new String[dependeyPathMap.size()]));
    }

    public static String[] getChildrenPath(long artifcatID, long versionNumber,
                                           Connection connection) throws SQLException {

        String selectSql = "SELECT DAID, DVN FROM DEPENDENCY WHERE AID=? AND VN=?";
        PreparedStatement s = connection.prepareStatement(selectSql);
        s.setLong(1, artifcatID);
        s.setLong(2, versionNumber);
        ResultSet result = s.executeQuery();

        ArrayList dependencyPathList = new ArrayList();

        //to see whether the patent node is there in the table , if not need to add that
        while (result.next()) {
            long daid = result.getLong("DAID");
            long childVersion = result.getLong("DVN");
            String SQL = "SELECT PATH FROM ARTIFACTS WHERE AID=?";
            PreparedStatement ps = connection.prepareStatement(SQL);
            ps.setLong(1, daid);
            ResultSet pathResult = ps.executeQuery();
            if (pathResult.next()) {
                String path = pathResult.getString("PATH") + "?v=" + childVersion;
                dependencyPathList.add(path);
            }
        }
        return (String[]) (dependencyPathList.toArray(new String[dependencyPathList.size()]));
    }

    public static String getParentPath(long artifcatID, long versionNumber,
                                           Connection connection) throws SQLException {

        if (versionNumber == -1) {
            versionNumber = getLatestVerison(artifcatID, connection);
        }
        if (versionNumber == -1) {
            return null;
        }
        String selectSql = "SELECT AID , VN FROM DEPENDENCY WHERE DAID=? AND DVN=?";
        PreparedStatement s = connection.prepareStatement(selectSql);
        s.setLong(1, artifcatID);
        s.setLong(2, versionNumber);
        ResultSet result = s.executeQuery();

        String parentPath = "";

        //to see whether the patent node is there in the table , if not need to add that
        if (result.next()) {
            long aid = result.getLong("AID");
            long parentVersionNumber = result.getLong("VN");
            String SQL = "SELECT PATH FROM ARTIFACTS WHERE AID=?";
            PreparedStatement ps = connection.prepareStatement(SQL);
            ps.setLong(1, aid);
            ResultSet pathResult = ps.executeQuery();
            if (pathResult.next()) {
                String path = pathResult.getString("PATH");
                parentPath = path + "?v=" + parentVersionNumber;
            }
        }
        return parentPath;
    }

    public static String[] getLatestChildrenPath(long artifcatID, long versionNumber,
                                           Connection connection) throws SQLException {

        String selectSql = "SELECT DAID, DVN FROM DEPENDENCY WHERE AID=? AND VN=?";
        PreparedStatement s = connection.prepareStatement(selectSql);
        s.setLong(1, artifcatID);
        s.setLong(2, versionNumber);
        ResultSet result = s.executeQuery();

        ArrayList dependencyPathList = new ArrayList();

        //to see whether the patent node is there in the table , if not need to add that
        while (result.next()) {
            long daid = result.getLong("DAID");
            String SQL = "SELECT PATH FROM ARTIFACTS WHERE AID=?";
            PreparedStatement ps = connection.prepareStatement(SQL);
            ps.setLong(1, daid);
            ResultSet pathResult = ps.executeQuery();
            if (pathResult.next()) {
                String path = pathResult.getString("PATH");
                dependencyPathList.add(path);
            }
        }
        return (String[]) (dependencyPathList.toArray(new String[dependencyPathList.size()]));
    }

    /**
     * To restoreVersion or rollback artifcat to a given version number
     *
     * @param artifactId : Resource ID
     * @param connection : Connection to DB
     * @param toVersion  : to which version we need to restoreVersion
     * @throws SQLException : If something went wrong
     */
    public static void revert(long artifactId, long toVersion, Connection connection) throws SQLException {

        String SQL = "SELECT CONTENT , AUTHOR ,UPDATED_TIME FROM VERSIONS WHERE AID=? AND VN=?";
        PreparedStatement statement = connection.prepareStatement(SQL);
        statement.setLong(1, artifactId);
        statement.setLong(2, toVersion);

        ResultSet result = statement.executeQuery();
        long versionNumber = 0;
        if (result.next()) {
            versionNumber = insertToVersionTable(connection, result.getObject(1),
                    artifactId,
                    result.getString(2),
                    result.getDate(3).getTime());
        }
        Resource resource = new ResourceDAO().getArtifactByID(artifactId, toVersion, connection);

        //updating max version table
        updateDependency(resource, connection, versionNumber, false);
        if (resource.isDirectory()) {
            String selectChildSql = "SELECT DAID, DVN FROM DEPENDENCY WHERE AID=? AND VN=?";
            PreparedStatement selectps = connection.prepareStatement(selectChildSql);
            selectps.setLong(1, artifactId);
            selectps.setLong(2, toVersion);
            ResultSet childResult = selectps.executeQuery();
            ArrayList dependencyList = new ArrayList();
            while (childResult.next()) {
                long daid = childResult.getLong("DAID");
                long dvn = childResult.getLong("DVN");
                PreparedStatement ps = getStatementForDependencyTable(connection,
                        artifactId, versionNumber, daid, dvn);
                dependencyList.add(ps);
            }

            for (int i = 0; i < dependencyList.size(); i++) {
                PreparedStatement preparedStatement = (PreparedStatement) dependencyList.get(i);
                preparedStatement.executeUpdate();
            }
        }

    }

    private static long insertToVersionTable(Connection connection,
                                             Object content,
                                             long id,
                                             String lastUpdaterUserName,
                                             long lastModified) throws SQLException {
        String sql = "INSERT INTO VERSIONS (AID, VN ,CONTENT , AUTHOR, UPDATED_TIME) VALUES (?,?,?,?,?)";
        PreparedStatement statement = connection.prepareStatement(sql);
        statement.setLong(1, id);
        //Need to getthe old version and increment that
        //Selecet VID from latest table ,
        // if it is there then find from that  , if it is not there then we need ti start from 0
        long versionNumber;
        versionNumber = getLatestVerison(id, connection);
        if (versionNumber == -1) {
            versionNumber = 1;
        } else {
            //incrementing the version number
            versionNumber = versionNumber + 1;
        }
        statement.setLong(2, versionNumber);
        if (content != null && content instanceof byte[]) {
            statement.setBytes(3, (byte[]) content);
        } else {
            statement.setBytes(3, null);
        }
        statement.setString(4, lastUpdaterUserName);
        statement.setTimestamp(5, new Timestamp(lastModified));
        statement.executeUpdate();
        return versionNumber;
    }

    /**
     * @param resource      : Artifcat we need to update itslef and its immediate parent
     * @param connection    : DB connetion
     * @param versionNumber : Version number of the artifcat
     * @param delete        : delete or adding
     * @return : Will retuen the resource version of the parent resource
     * @throws SQLException : if something went wrong
     */
    private static long updateDependencyTable(Resource resource,
                                              Connection connection,
                                              long versionNumber,
                                              boolean delete) throws SQLException {
        /**
         *        D0
         *    D1      D2
         *  D3  R1    R2
         *
         *  First fine D1 V0
         */
        //Checking whether the path is null or top most directory
        // Checking whether resource is root node or not
        String path = resource.getPath();
        //Checking whether the path is null or top most directory
        //if (path != null && path.indexOf("/") > -1) {
        if (path != null && !path.equals(RegistryConstants.ROOT_PATH)) {
            int lastIndex = path.lastIndexOf("/");
            String parentPath = RegistryConstants.ROOT_PATH;
            if (lastIndex != 0) {
                parentPath = path.substring(0, lastIndex);                
            }

            if ("".equals(parentPath)) {
//TODO : Need to think the logic here
            } else {
                ResourceDAO resourceDAO = new ResourceDAO();
                long parentId = resourceDAO.getArtifactID(parentPath, connection);
                if (resource.getLastModified() == null) {
                    resource.setLastModified(new java.util.Date());
                }
                // Need to get the parent's parent and update the and update the

                long artifactversionNumber = getLatestVerison(parentId, connection);

                long nextVersionNumber = getNextVersionNumber(parentId,
                        resource.getLastUpdaterUserName(),
                        connection, resource.getLastModified().getTime());
                String selectSql = "SELECT DAID , DVN FROM DEPENDENCY WHERE AID=? AND VN=?";
                PreparedStatement s = connection.prepareStatement(selectSql);
                s.setLong(1, parentId);
                s.setLong(2, artifactversionNumber);
                ResultSet result = s.executeQuery();

                ArrayList dependeyMap = new ArrayList();

                //to see whether the patent node is there in the table , if not need to add that
                while (result.next()) {
                    long daid = result.getLong("DAID");
                    if (daid != resource.getId()) {
                        long dvn = result.getLong("DVN");
                        PreparedStatement ps = getStatementForDependencyTable(connection,
                                parentId, nextVersionNumber, daid, dvn);
                        dependeyMap.add(ps);
                    }
                }
                if (!delete) {
                    PreparedStatement ds = getStatementForDependencyTable(connection,
                            parentId, nextVersionNumber, resource.getId(), versionNumber);
                    dependeyMap.add(ds);
                }
                for (int i = 0; i < dependeyMap.size(); i++) {
                    PreparedStatement preparedStatement = (PreparedStatement) dependeyMap.get(i);
                    preparedStatement.executeUpdate();
                }
                return nextVersionNumber;
            }
        } else {
            //TODO , I think we do not need to add the entry here , since when adding a child node,
            //  if the parent is not there then we will add that
        }
        return -1;
    }

    private static PreparedStatement getStatementForDependencyTable(Connection connection,
                                                                    long parentId,
                                                                    long nextVersionNumber,
                                                                    long daid,
                                                                    long dvn) throws SQLException {
        String SQL = "INSERT INTO DEPENDENCY (AID, VN , DAID , DVN) VALUES (?,? ,?,?)";
        PreparedStatement ps = connection.prepareStatement(SQL);
        ps.setLong(1, parentId);
        ps.setLong(2, nextVersionNumber);
        ps.setLong(3, daid);
        ps.setLong(4, dvn);
        System.out.println("Adding dependency: " + parentId + " " + nextVersionNumber + " " + daid + " " + dvn);
        return ps;
    }

    private static long getNextVersionNumber(long artifactID,
                                             String lastUpdaterUserName,
                                             Connection connection,
                                             long lastModifiedTime) throws SQLException {
        long maxVersion = getLatestVerison(artifactID, connection);
        if (maxVersion == -1) {
            maxVersion = 0;
        }
        //here maxVersion is always >0
        maxVersion = maxVersion + 1;
        insertToVersionTable(connection, null, artifactID, lastUpdaterUserName, lastModifiedTime);
        return maxVersion;
    }

}
